package cz.fidentis.analyst.comparison;

import cz.fidentis.analyst.mesh.core.MeshFacet;
import cz.fidentis.analyst.mesh.core.MeshPoint;
import javafx.util.Pair;

import javax.vecmath.Vector3d;
import java.util.ArrayList;
import java.util.Comparator;
import java.util.List;
import java.util.Optional;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;
import java.util.stream.Collectors;

public class HausdorffDistance {
    private MeshFacet firstFacet;
    private MeshFacet secondFacet;

    public HausdorffDistance(MeshFacet firstFacet, MeshFacet secondFacet) {
        this.firstFacet = firstFacet;
        this.secondFacet = secondFacet;
    }


    /**
     * finds the nearest vertex on the second facet
     *
     * @param vertex vertex from
     * @return vertex, nearest vertex from second facet, distance
     */
    private ClosestVertices getNearestVertex(MeshPoint vertex) {
        /*
        double minDistance = Double.MAX_VALUE;
        MeshPoint closestVertex = null;
        for (MeshPoint comparedVertex :
                facet.getVertices()) {
            double distance = getDistanceBetweenPoints(vertex.getPosition(), comparedVertex.getPosition());
            if (minDistance > distance) {
                minDistance = distance;
                closestVertex = comparedVertex;
            }
        }
        return new ClosestVertices(vertex, closestVertex, minDistance);
        */

        Vector3d position = vertex.getPosition();
        Optional<Pair<MeshPoint, Double>> closestVertexAndDistance = secondFacet.getVertices().parallelStream()
                .map((meshPoint) -> new Pair<>(meshPoint, getDistanceBetweenPoints(position, meshPoint.getPosition())))
                .max((Comparator.comparingDouble(Pair::getValue)));
        return closestVertexAndDistance.map(vector3dDoublePair -> new ClosestVertices(vertex,
                vector3dDoublePair.getKey(),
                vector3dDoublePair.getValue())).orElse(null);
    }

    /**
     * returns distance between two points
     *
     * @param point1 first point
     * @param point2 second point
     * @return distance
     */
    private double getDistanceBetweenPoints(Vector3d point1, Vector3d point2) {
        Vector3d helperVector = new Vector3d();
        helperVector.sub(point1, point2);
        return helperVector.length();
    }

    /**
     * calculates Hausdorff Distance to the nearest vertex from second facet for each vertex in first facet
     *
     * @return list containing vertex from first facet, closest vertex to it from second facet, distance
     */
    public List<ClosestVertices> calculateHausdorffDistanceToVertices() {
        int numberOfVertices = firstFacet.getNumberOfVertexes();
        List<Future<ClosestVertices>> closestVerticesFutures = new ArrayList<>(numberOfVertices);
        ExecutorService executor = Executors.newCachedThreadPool();
        for (final MeshPoint vertex :
                firstFacet.getVertices()) {
            closestVerticesFutures.add(executor.submit(() -> getNearestVertex(vertex)));
        }
        List<ClosestVertices> closestVertices = new ArrayList<>(numberOfVertices);
        for (Future<ClosestVertices> future :
                closestVerticesFutures) {
            executor.submit(() -> {
                while (true) {
                    if (future.isDone()) {
                        try {
                            closestVertices.add(future.get());
                        } catch (InterruptedException | ExecutionException e) {
                            e.printStackTrace();
                        }
                        break;
                    }
                }
            });
        }
        executor.shutdown();
        return closestVertices;
    }

    /**
     * @param facet         a
     * @param comparedFacet a
     * @return a
     */
    public List<ClosestVertices> calculateHausdorffDistanceToVertices2
    (MeshFacet facet, MeshFacet comparedFacet) {
        return facet.getVertices().parallelStream()
                .map(this::getNearestVertex)
                .collect(Collectors.toList());
    }

    /**
     * calculates Hausdorff Distance to the nearest point on second facet for each vertex
     * @return list containing vertex from first facet, closest point to it from second facet, distance
     */
    public List<ClosestVertices> calculateHausdorffDistanceToMesh() {
        return firstFacet.getVertices().parallelStream()
                .map((meshPoint) ->
                        CalculateNearestPointOnMesh(meshPoint,
                                secondFacet.getCornerTable().getTriangleIndexesByVertexIndex(
                                        secondFacet.getVertices().indexOf(
                                                getNearestVertex(meshPoint).getSecondVertex())))
                ).collect(Collectors.toList());

    }

    /**
     * calculates Hausdorff Distance to the nearest point on second facet for vertex
     * @param vertex vertex from first facet
     * @param indicesOfTrianglesOfVertex indices of triangles that contain the nearest vertex on second mesh
     * @return vertex from first facet, closest point to it from second facet, distance
     */
    private ClosestVertices CalculateNearestPointOnMesh(
            MeshPoint vertex, List<Integer> indicesOfTrianglesOfVertex) {
        Vector3d vertexPosition = vertex.getPosition();
        List<Pair<Vector3d, Double>> projections = new ArrayList<>(indicesOfTrianglesOfVertex.size());
        Vector3d helperVector = new Vector3d();
        for (int index :
                indicesOfTrianglesOfVertex) {
            List<Vector3d> triangle = secondFacet.getVerticesOfTriangle(index).stream()
                    .map(MeshPoint::getPosition)
                    .collect(Collectors.toList());
            Vector3d projection = getProjectionToTrianglePlane(vertexPosition, triangle);
            if (isPointInTriangle(projection, triangle)) {
                helperVector.sub(vertexPosition, projection);
                projections.add(new Pair<>(projection, helperVector.length()));
            } else {
                projection = getProjectionToClosestEdge(projection, triangle);
                helperVector.sub(vertexPosition, projection);
                projections.add(new Pair<>(projection, helperVector.length()));
            }

        }

        Pair<Vector3d, Double> closestPosition = projections.stream()
                .min(Comparator.comparingDouble(Pair::getValue)).orElseGet(() -> new Pair<>(null, Double.MAX_VALUE));
        return new ClosestVertices(
                vertex,
                new MeshPoint(closestPosition.getKey(), null, null),
                closestPosition.getValue());
    }

    /**
     * returns perpendicular projection from vertex to plane of triangle
     * @param vertex vertex from which the projection is created
     * @param triangle triangle that defines the plane
     * @return projection to plane of triangle
     */
    private Vector3d getProjectionToTrianglePlane(Vector3d vertex, List<Vector3d> triangle) {
        Vector3d ab = new Vector3d();
        ab.sub(triangle.get(0), triangle.get(1));
        Vector3d ac = new Vector3d();
        ac.sub(triangle.get(0), triangle.get(2));
        Vector3d normal = new Vector3d();
        normal.cross(ab, ac);
        normal.normalize();

        Vector3d helperVector = new Vector3d(vertex);
        helperVector.sub(triangle.get(0));
        double distance = helperVector.dot(normal);
        helperVector.scaleAdd(-distance, normal, helperVector);
        return helperVector;
    }

    /**
     * checks if a point in plane of triangle lies within the triangle
     * @param point checked point
     * @param triangle triangle
     * @return true if point is in triangle, false otherwise
     */
    private boolean isPointInTriangle(Vector3d point, List<Vector3d> triangle) {
        Vector3d p0 = triangle.get(0);
        Vector3d p1 = triangle.get(1);
        Vector3d p2 = triangle.get(2);

        Vector3d u = new Vector3d(p1.x - p0.x, //T1-T0
                p1.y - p0.y,
                p1.z - p0.z);
        Vector3d v = new Vector3d(p2.x - p0.x,//T2-T0
                p2.y - p0.y,
                p2.z - p0.z);

        Vector3d w = new Vector3d(point.x - p0.x, //PI-T0
                point.y - p0.y,
                point.z - p0.z);

        double uDotV = u.dot(v);
        double uDotU = u.dot(u);
        double vDotV = v.dot(v);
        double wDotU = w.dot(u);
        double wDotV = w.dot(v);
        double s = (uDotV * wDotV - vDotV * wDotU) / (uDotV * uDotV - uDotU * vDotV);
        double t = (uDotV * wDotU - uDotU * wDotV) / (uDotV * uDotV - uDotU * vDotV);

        return s >= 0 && t >= 0 && s + t <= 1;
    }

    /**
     checks if a point in plane of triangle lies within the triangle
     * @param point checked point
     * @param triangle triangle
     * @return true if point is in triangle, false otherwise
     */
    private boolean isPointInTriangle2(Vector3d point, List<Vector3d> triangle) {
        List<Vector3d> pointToVertices = triangle.stream()
                .map((vertex) -> {
                    Vector3d v = new Vector3d(vertex);
                    v.sub(point);
                    return v;
                })
                .collect(Collectors.toList());
        double angleSum = 0;
        for (int i = 0; i < 3; i++) {
            angleSum += pointToVertices.get(i).angle(pointToVertices.get((i + 1) % 3));
        }
        angleSum -= Math.PI;
        return -0.01 < angleSum && angleSum < 0.01;
    }

    /**
     * returns projection to to the nearest edge of triangle
     * @param point point in plane of triangle
     * @param triangle triangle
     * @return perpendicular projection to the nearest edge
     */
    private Vector3d getProjectionToClosestEdge(Vector3d point, List<Vector3d> triangle) {

        Vector3d[] projections = new Vector3d[3];
        projections[0] = getProjectionToEdge(point, triangle.get(0), triangle.get(1));
        projections[1] = getProjectionToEdge(point, triangle.get(1), triangle.get(2));
        projections[2] = getProjectionToEdge(point, triangle.get(2), triangle.get(0));
        double minDistance = Double.MAX_VALUE;
        Vector3d closestProjection = null;
        Vector3d helperVector = new Vector3d();
        for (Vector3d projection :
                projections) {
            helperVector.sub(point, projection);
            double distance = helperVector.length();
            if (distance < minDistance) {
                minDistance = distance;
                closestProjection = projection;
            }
        }
        return closestProjection;
    }

    /**
     * returns projection to edge
     * @param point point in plane of triangle
     * @param edgeVertex1 first vertex of edge
     * @param edgeVertex2 second vertex of edge
     * @return projection to edge
     */
    private Vector3d getProjectionToEdge(Vector3d point, Vector3d edgeVertex1, Vector3d edgeVertex2) {
        Vector3d ab = new Vector3d();
        ab.sub(edgeVertex1, edgeVertex2);
        Vector3d ap = new Vector3d();
        ap.sub(edgeVertex1, point);
        double t = ab.dot(ap) / ab.lengthSquared();
        return new Vector3d(edgeVertex1.x + t * ab.x, edgeVertex1.y + t * ab.y, edgeVertex1.z + t * ab.z);
    }
}
