package cz.fidentis.analyst.mesh.core;

import javax.vecmath.Vector3d;

/**
 * @author Matej Lukeš
 */
public class MeshPoint {
    private Vector3d position, normal, texCoord;

    /**
     * Constructor of MeshPoint
     *
     * @param position position of MeshPoint
     * @param normal   normal of MeshPoint
     * @param texCoord texture coordinates of MeshPoint
     */
    public MeshPoint(Vector3d position, Vector3d normal, Vector3d texCoord) {
        if (position == null) {
            throw new NullPointerException("position cannot be null");
        }

        this.position = position;
        this.normal = normal;
        this.texCoord = texCoord;
    }

    /**
     * copy constructor of meshPoint
     *
     * @param meshPoint copied meshPoint
     */
    public MeshPoint(MeshPoint meshPoint) {
        this.position = new Vector3d(meshPoint.position);
        if (meshPoint.normal != null) {
            this.normal = new Vector3d(meshPoint.normal);
        }
        if (meshPoint.texCoord != null) {
            this.texCoord = new Vector3d(meshPoint.texCoord);
        }
    }

    /**
     * @return normal
     */
    public Vector3d getNormal() {
        return new Vector3d(normal);
    }

    /**
     * @return position
     */
    public Vector3d getPosition() {
        return new Vector3d(position);
    }

    /**
     * @return texture coordinates
     */
    public Vector3d getTexCoord() {
        return new Vector3d(texCoord);
    }

    /**
     * @param obj compared object
     * @return true if positions, normals and texture coordinates are equal, false otherwise
     */
    @Override
    public boolean equals(Object obj) {
        if (!(obj instanceof MeshPoint)) {
            return false;
        }

        MeshPoint meshPointObj = (MeshPoint) obj;
        return this.position.equals(meshPointObj.position);
    }

    @Override
    public int hashCode() {
        return position.hashCode();
    }

    /**
     * returns new instance of MeshPoint with subtracted position
     *
     * @param subtrahend position to be subtracted
     * @return subtracted MeshPoint
     */
    public MeshPoint subtractPosition(MeshPoint subtrahend) {
        return MeshPoint.this.subtractPosition(subtrahend.position);
    }

    /**
     * returns new instance of MeshPoint with subtracted position
     *
     * @param subtrahend position to be subtracted
     * @return subtracted MeshPoint
     */
    public MeshPoint subtractPosition(Vector3d subtrahend) {
        Vector3d newPosition = new Vector3d(position);
        newPosition.sub(subtrahend);
        if (normal != null) {
            if (texCoord != null) {
                return new MeshPoint(new Vector3d(newPosition), new Vector3d(normal), new Vector3d(texCoord));
            }
            return new MeshPoint(new Vector3d(newPosition), new Vector3d(normal), null);
        }
        return new MeshPoint(new Vector3d(newPosition), null, null);
    }

    /**
     * returns new instance of MeshPoint with added position
     *
     * @param addend position to be added
     * @return added MeshPoint
     */
    public MeshPoint addPosition(MeshPoint addend) {
        return MeshPoint.this.addPosition(addend.position);
    }

    /**
     * returns new instance of MeshPoint with added position
     *
     * @param addend position to be added
     * @return added MeshPoint
     */
    public MeshPoint addPosition(Vector3d addend) {
        Vector3d newPosition = new Vector3d(position);
        newPosition.add(addend);
        if (normal != null) {
            if (texCoord != null) {
                return new MeshPoint(new Vector3d(newPosition), new Vector3d(normal), new Vector3d(texCoord));
            }
            return new MeshPoint(new Vector3d(newPosition), new Vector3d(normal), null);
        }
        return new MeshPoint(new Vector3d(newPosition), null, null);
    }

    /**
     * returns new instance of MeshPoint with multiplied position by number
     *
     * @param number Number for multiplying
     * @return multiplied MeshPoint
     */
    public MeshPoint multiplyPosition(double number) {
        if (normal != null) {
            if (texCoord != null) {
                return new MeshPoint(new Vector3d(this.getPosition().x * number,
                        this.getPosition().y * number, this.getPosition().z * number),
                        new Vector3d(normal), new Vector3d(texCoord));
            }
            return new MeshPoint(new Vector3d(this.getPosition().x * number,
                    this.getPosition().y * number, this.getPosition().z * number),
                    new Vector3d(normal), null);
        }
        return new MeshPoint(new Vector3d(this.getPosition().x * number,
                this.getPosition().y * number, this.getPosition().z * number),
                null, null);
    }

    /**
     * returns new instance of MeshPoint with divided position by number
     *
     * @param number Number for division
     * @return divided MeshPoint
     */
    public MeshPoint dividePosition(double number) {
        if (normal != null) {
            if (texCoord != null) {
                return new MeshPoint(new Vector3d(this.getPosition().x / number, this.getPosition().y / number,
                        this.getPosition().z / number), new Vector3d(normal), new Vector3d(texCoord));
            }
            return new MeshPoint(new Vector3d(this.getPosition().x / number, this.getPosition().y / number,
                    this.getPosition().z / number), new Vector3d(normal), null);
        }
        return new MeshPoint(new Vector3d(this.getPosition().x / number, this.getPosition().y / number,
                this.getPosition().z / number), null, null);
    }

    /**
     * Returns the cross product of two points.
     *
     * @param meshPoint Second argument of the cross product operation.
     * @return MeshPoint representing the resulting vector.
     */
    public MeshPoint crossProduct(MeshPoint meshPoint) {
        Vector3d newPosition = new Vector3d();
        newPosition.cross(position, meshPoint.position);

        return new MeshPoint(newPosition, new Vector3d(normal), new Vector3d(texCoord));
    }

    /**
     * returns the dot product of two points
     *
     * @param meshPoint Second argument of the dot product operation
     * @return dot product of two instances of MeshPoint
     */
    public double dotProduct(MeshPoint meshPoint) {
        return position.dot(meshPoint.position);
    }

    /**
     * returns the dot product of two points
     *
     * @param vector Second argument of the dot product operation
     * @return dot product of two instances of MeshPoint
     */
    public double dotProduct(Vector3d vector) {
        return position.dot(vector);
    }


    /**
     * returns absolute value of MeshPoint
     *
     * @return absolute value of MeshPoint
     */
    public double abs() {
        return Math.sqrt(this.getPosition().x * this.getPosition().x +
                this.getPosition().y * this.getPosition().y + this.getPosition().z * this.getPosition().z);
    }
}
